#include <iostream>
#include <thread>
#include <atomic>
#include <chrono>
#include <set>
#include <vector>
#include <string.h>
#include <tt_cplus_sdk.h>
#include "enums/OrderSide.h"
#include "enums/TimeInForce.h"
#include "enums/OrderType.h"
#include "enums/ExecType.h"
#include "enums/ProductType.h"
#include <iomanip>
#include <sstream>
#include <condition_variable>
#include "shared_ptr.h"


std::mutex mutex;
std::condition_variable sdkReadyCondition;
std::condition_variable releaseReady;
std::condition_variable reserveReady;
#define ACCOUNT_ID 00000

class RouteOrder : public ttsdk::IOrderEventHandler,
    public ttsdk::IPriceEventHandler
{
public:
    RouteOrder(ttsdk::InstrumentPtr instrument)
        : instrument_(instrument), running_(false), shuttingDown_(false)
    {
    };

    virtual ~RouteOrder()
    {
        Stop();
    };

    void Stop()
    {
        if (shuttingDown_) return;

        shuttingDown_ = true;
        if (running_)
        {
            
            // cancel order if working
            SendCancel();
        }
        else
        {
            ttsdk::UnsubscribePrices(subId_);
        }

    };

    void Start()
    {
        ttsdk::TTLogInfo("Starting Position Reserve strategy");
        ttsdk::Instrument::ResponseCode respCode;
        if (!instrument_)
        {
            std::cout << "Unable to get instrument . Stopping strategy." << std::endl;
            return;
        }

        std::cout << "Subscribing for prices for instrument : "  << instrument_->GetAlias() << std::endl;
        subId_ = ttsdk::SubscribePrices(instrument_, 10, true, true, false, this);
    }

    virtual void OnUnsubscribed(const uint64_t subId)
    {
        running_ = false;
        std::cout << "OnUnsubscribed from prices. subId=" << subId << std::endl;
    };
    virtual void OnDisconnect(const uint64_t subId, const ttsdk::InstrumentPtr&, const ttsdk::PriceEventType)
    {
        std::cout << "OnDisconnected from prices. subId=" << subId << std::endl;
    };
    virtual void OnPriceUpdate(const uint64_t, const ttsdk::InstrumentPtr&, const ttsdk::PriceSnap& snap)
    {
        double price = 0.0;

        if (snap.bids.count > 0)
            price = snap.bids.levels[0].price;
        else if (!std::isnan(snap.last_trade_price))
            price = snap.last_trade_price;
        else if (!std::isnan(snap.settlement_price))
            price = snap.settlement_price;
           
        running_ = true;

        if (!order1_)
        {
            order1_ = ttsdk::CreateOrderFromPriceSub(subId_);
            if (!order1_)
            {
                std::cout << "Order 1 creation failed. Stopping strategy." << std::endl;
                return;
            }
            order1_->Subscribe(*this);

            std::cout << "Placing order 1" << std::endl;

            profile1_.request_id++;
            profile1_.account_id = ACCOUNT_ID;
            profile1_.side = ttsdk::OrderSide::Buy;
            profile1_.price = instrument_->TickPriceUp(price, 1);
            profile1_.quantity = 5;
            profile1_.type = ttsdk::OrderType::Limit;
            profile1_.tif = ttsdk::TimeInForce::Day;

            order1_->SendNew(profile1_);
        }

        if (!order2_)
        {
            order2_ = ttsdk::CreateOrderFromPriceSub(subId_);
            if (!order2_)
            {
                std::cout << "Order2 creation failed. Stopping strategy." << std::endl;
                return;
            }
            order2_->Subscribe(*this);

            std::cout << "Placing order 2" << std::endl;

            profile2_.request_id++;
            profile2_.account_id = ACCOUNT_ID;
            profile2_.side = ttsdk::OrderSide::Buy;
            profile2_.price = instrument_->TickPriceUp(price, 2);
            profile2_.quantity = 5;
            profile2_.type = ttsdk::OrderType::Limit;
            profile2_.tif = ttsdk::TimeInForce::Day;

            order2_->SendNew(profile2_);
        }

        if (!order3_)
        {
            order3_ = ttsdk::CreateOrderFromPriceSub(subId_);
            if (!order3_)
            {
                std::cout << "Order3 creation failed. Stopping strategy." << std::endl;
                return;
            }
            order3_->Subscribe(*this);

            std::cout << "Placing order 3" << std::endl;

            profile3_.request_id++;
            profile3_.account_id = ACCOUNT_ID;
            profile3_.side = ttsdk::OrderSide::Buy;
            profile3_.price = instrument_->TickPriceUp(price, 3);
            profile3_.quantity = 5;
            profile3_.type = ttsdk::OrderType::Limit;
            profile3_.tif = ttsdk::TimeInForce::Day;

            order3_->SendNew(profile3_);
        }
    };

    virtual void OnError(const uint64_t subId, const ttsdk::InstrumentPtr& instrument, const ttsdk::SubscriptionError code, const char* error)
    {
        std::cout << "Error occurred subscribing for prices for instr=" << instrument->GetName()
            << " code=" << (int)code << " error=" << error << " subId=" << subId << std::endl;
    };
   
    void SendCancel()
    {
        if (order1_)
        {
            auto execRpt1 = order1_->GetCurrentState();
            order1_->Unsubscribe();
            if (execRpt1 && execRpt1->IsWorking())
            {
                profile1_.request_id++;
                order1_->SendCancel(profile1_);
            }
            
        }

        if (order2_)
        {
            // cancel order if working
            auto execRpt2 = order2_->GetCurrentState();
            order2_->Unsubscribe();
            if (execRpt2 && execRpt2->IsWorking())
            {
                profile2_.request_id++;
                order2_->SendCancel(profile2_);
            }
            
        }
        
        if (order3_)
        {
            // cancel order if working
            auto execRpt3 = order3_->GetCurrentState();
            order3_->Unsubscribe();
            if (execRpt3 && execRpt3->IsWorking())
            {
                profile3_.request_id++;
                order3_->SendCancel(profile3_);
            }
        }
    }

    virtual void OnUnsubscribed(const char*) override
    {
        ttsdk::TTLogInfo("Order event handler unsubscribed");
        ttsdk::UnsubscribePrices(subId_);

        order1_.reset();
        order2_.reset();
        order3_.reset();

        instrument_.reset();
    }

    virtual void OnExecutionReport(ttsdk::OrderPtr order, ttsdk::ExecutionReportPtr execRpt) override
    {
        switch (execRpt->GetExecType())
        {
        case ttsdk::ExecType::New:
            std::cout << "Order successfully entered in market. RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            break;
        case ttsdk::ExecType::Canceled:
            std::cout << "Order successfully canceled. RequestId=" << execRpt->GetRequestId() <<
                " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << " [external=" << execRpt->IsExternalAction() <<
                "]" << std::endl;

            if (order1_ && order->GetOrderId() == order1_->GetOrderId())
                order1_.reset();

            if (order2_ && order->GetOrderId() == order2_->GetOrderId())
                order2_.reset();

            if (order3_ && order->GetOrderId() == order3_->GetOrderId())
                order3_.reset();

            break;
        case ttsdk::ExecType::Replaced:
            std::cout << "Order replaced. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) <<
                " [external=" << execRpt->IsExternalAction() << "]" << std::endl;
            break;
        case ttsdk::ExecType::Trade:
            std::cout << "Order filled. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            // clear the order variable so a new order will be placed
            
            if (order1_ && (order->GetOrderId() == order1_->GetOrderId()))
                order1_.reset();

            if (order2_ && order->GetOrderId() == order2_->GetOrderId())
                order2_.reset();

            if (order3_ && order->GetOrderId() == order3_->GetOrderId())
                order3_.reset();

            break;
        case ttsdk::ExecType::Rejected:
            std::cout << "Order filled. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;
            break;
        default:
            std::cout << "Unhandled event. RequestId=" << execRpt->GetRequestId() <<
                " TTOrderId=" << order->GetOrderId() << " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) <<
                " ExecutionType=" << ttsdk::ToString(execRpt->GetExecType()) << std::endl;
        }
    }
    virtual void OnReject(ttsdk::OrderPtr order, ttsdk::RejectResponsePtr rejResp) override
    {
        std::cout << "OnReject event. RequestId=" << rejResp->GetRequestId() <<
            " TTOrderId=" << order->GetOrderId() << " RejectReason=" << ttsdk::ToString(rejResp->GetRejectReason()) <<
            " Text=" << rejResp->GetText() << std::endl;
        if (order->GetOrderId() == order1_->GetOrderId())
            order1_.reset();

        if (order->GetOrderId() == order2_->GetOrderId())
            order2_.reset();

        if (order->GetOrderId() == order3_->GetOrderId())
            order3_.reset();
    }
    virtual void OnSendFailed(ttsdk::OrderPtr order, const ttsdk::OrderProfile& profile, const ttsdk::IOrderEventHandler::SendCode code) override
    {
        if (code == ttsdk::IOrderBookEventHandler::SendCode::EXCEEDED_PREALLOCATED_RISK || code == ttsdk::IOrderBookEventHandler::SendCode::EXCEEDED_PREALLOCATED_RISK_CLIP_SIZE)
        {
            //Risk Rejection processing should be done here. 
            std::cout << "Order send failed due to risk rejection, RequestId=" << profile.request_id << " TTOrderId=" << order->GetOrderId() << std::endl;
        }
        else
            std::cout << "Order send failed. RequestId=" << profile.request_id << " TTOrderId=" << order->GetOrderId() << " SendCode=" << (int)code << std::endl;
    }

    ttsdk::OrderProfile profile1_;
    ttsdk::OrderProfile profile2_;
    ttsdk::OrderProfile profile3_;
    ttsdk::InstrumentPtr instrument_;
    ttsdk::OrderPtr order1_;
    ttsdk::OrderPtr order2_;
    ttsdk::OrderPtr order3_;
    std::atomic<bool> running_;
    std::atomic<bool> shuttingDown_;
    uint64_t subId_;
};


class MyOrderBookHandler : public ttsdk::IOrderBookEventHandler
{
public:
    MyOrderBookHandler() : IOrderBookEventHandler() {};
    virtual ~MyOrderBookHandler() {};
    virtual void OnExecutionReport(ttsdk::OrderPtr order, ttsdk::ExecutionReportPtr execRpt) override
    {
        std::cout << "MyOrderBookHandler::OnExecutionReport TTOrderId=" << order->GetOrderId() <<
            " OrderStatus=" << ttsdk::ToString(execRpt->GetOrderStatus()) << std::endl;

    };
    virtual void OnEpiqUpdate(ttsdk::OrderPtr order, double epiq) override
    {
        std::string order_id = order->GetOrderId();
        double e = epiq;
        std::cout << "\nOnEpiqUpdate::orderId=" << order_id << " EPIQ=" << e << std::endl;
    }
    virtual void OnReject(ttsdk::OrderPtr order, ttsdk::RejectResponsePtr rejResp) override
    {
    }
    virtual void OnSendFailed(ttsdk::OrderPtr order, const ttsdk::OrderProfile& profile, const SendCode code) override
    {
    }
    virtual void OnUnsubscribed(const char* orderId) override
    {
    }
    virtual void OnAccountDownloadEnd(const uint64_t accountId) override
    {
    };
    virtual void OnOrderBookDownloadEnd() override
    {
        std::cout << "OnOrderBookDownloadEnd -- orders are ready to use" << std::endl;
        sdkReadyCondition.notify_one();
    };
    virtual void OnRiskReserved(const uint64_t instrumentId, const uint64_t accountId, const ttsdk::RiskSide side, const bool successful)
    {
        if (!successful)
        {
            std::cout << " Failed to reserve risk for instrumentId : " << instrumentId << " accountId : " << accountId << " side :  " << ttsdk::ToString(side) << std::endl;
        }
        else
        {
            std::cout << " Reserved risk for instrumentId : " << instrumentId << " accountId : " << accountId << " side :  " << ttsdk::ToString(side) << std::endl;
            reserveReady.notify_one();
        }
    };
    virtual void OnRiskReleased(const uint64_t instrumentId, const uint64_t accountId, const ttsdk::RiskSide side, const bool successful)
    {
        if (!successful)
        {
            std::cout << " Failed to release risk for instrumentId : " << instrumentId << " accountId : " << accountId << " side :  " << ttsdk::ToString(side) << std::endl;
        }
        else
        {
            std::cout << " Released risk for instrumentId : " << instrumentId << " accountId : " << accountId << " side :  " << ttsdk::ToString(side) << std::endl;
            releaseReady.notify_one();
        }
    };
};

class SDKEventHandler : public ttsdk::IEventHandler
{
public:
    SDKEventHandler() : IEventHandler() {};
    virtual ~SDKEventHandler() {};
    virtual void OnStatus(const ttsdk::IEventHandler::Status status)
    {
        std::cout << "SDKEventHandler::OnStatus (" << (uint32_t)status << ")" << std::endl;
    };
};

class PositionManager {
public:
    PositionManager()
    {
    }

    virtual ~PositionManager()
    {
        ReleaseAllRisk();
    }

    void ReleaseAllRisk()
    {  
        for (auto itr = positionReserves_.cbegin(); itr != positionReserves_.cend();)
        {
            if (!ttsdk::ReleaseRisk(itr->first, ACCOUNT_ID, itr->second))
                std::cout << " Failed to release risk for instrumentId : " << itr->first->GetInstrumentId() << " accountId : " << ACCOUNT_ID << " side :  " << ttsdk::ToString(itr->second) << std::endl;
            
            // wait for the risk released event
            std::unique_lock<std::mutex> release_lock(releaseMutex);
            {
                if (releaseReady.wait_for(release_lock, std::chrono::seconds(300)) == std::cv_status::timeout)
                    std::cout << "Timeout waiting for the risk released event!" << std::endl;
                else
                    positionReserves_.erase(itr++);
            }
        }
    }

    bool ReserveRisk(ttsdk::InstrumentPtr instrument, ttsdk::RiskSide side, const double quantity, const double maxClipSize)
    {
        if (!ttsdk::ReserveRisk(instrument, ACCOUNT_ID, side, quantity, maxClipSize))
        {
            std::cout << " Failed to reserve risk for instrumentId : " << instrument->GetAlias() << " accountId : " << ACCOUNT_ID << " side :  " << ttsdk::ToString(side) << std::endl;
            return false;
        }

        // wait for the risk reserved event
        std::unique_lock<std::mutex> reserve_lock(reserveMutex);
        {
            if (reserveReady.wait_for(reserve_lock, std::chrono::seconds(300)) == std::cv_status::timeout)
            {
                std::cout << "Timeout waiting for the risk reserved event!" << std::endl;
                return false;
            }

            positionReserves_.insert(std::make_pair(instrument, side));
        }

        return true;
    }

    bool ReleaseRisk(ttsdk::InstrumentPtr instrument, ttsdk::RiskSide side)
    {
        if (!ttsdk::ReleaseRisk(instrument, ACCOUNT_ID, side))
        {
            std::cout << " Failed to release risk for instrumentId : " << instrument << " accountId : " << ACCOUNT_ID << " side :  " << ttsdk::ToString(side) << std::endl;
            return false;
        }

        // wait for the risk released event
        std::unique_lock<std::mutex> release_lock(releaseMutex);
        {
            if (releaseReady.wait_for(release_lock, std::chrono::seconds(300)) == std::cv_status::timeout)
            {
                std::cout << "Timeout waiting for the risk released event!" << std::endl;
                return false;
            }

            positionReserves_.erase(std::make_pair(instrument, side));
        }

        return true;
    }

    mutable std::mutex reserveMutex;
    mutable std::mutex releaseMutex;
    std::set<std::pair<ttsdk::InstrumentPtr, ttsdk::RiskSide>> positionReserves_;
};



int main(int argc, char* argv[])
{
    // this example is loading the app key secret from the command line or environment rather than hard coding.
    ttsdk::Environment env = ttsdk::Environment::UatCert;
    std::string sEnv("UatCert");

    std::string app_key;
    for (int i = 1; i < argc; ++i)
    {
        std::string arg = argv[i];
        if ((arg == "-k") || (arg == "--key"))
        {
            if (i + 1 < argc)
            {
                i++;
                app_key = argv[i];
            }
            else
            {
                std::cerr << "--key option requires one argument." << std::endl;
                return 1;
            }
        }
        else if ((arg == "-e") || (arg == "--env"))
        {
            if (i + 1 < argc)
            {
                i++;
                sEnv = argv[i];
                if (sEnv == "ProdLive") env = ttsdk::Environment::ProdLive;
                else if (sEnv == "ProdSim") env = ttsdk::Environment::ProdSim;
                else if (sEnv == "UatCert") env = ttsdk::Environment::UatCert;
                else if (sEnv == "DevCert")env = ttsdk::Environment::DevCert;
                else {
                    std::cerr << "Invalid env argument: " << sEnv << std::endl;
                    return 1;
                }
            }
            else
            {
                std::cerr << "--env option requires one argument." << std::endl;
                return 1;
            }
        }
    }
    if (app_key.empty())
    {
        auto env_app_key = std::getenv("TT_APP_KEY");
        if (!env_app_key || strcmp(env_app_key, "") == 0)
        {
            std::cout << "app_key is invalid. " << std::endl;
            return 1;
        }
        app_key = env_app_key;
    }
    std::cout << "using app_key: " << app_key << " in " << sEnv << std::endl;


    // these are the options for the SDK
    ttsdk::TTSDKOptions options;

    // set the environment the app needs to run (where your machine is located)
    options.environment = env;

    // set your app secret Key here. It looks like: 00000000-0000-0000-0000-000000000000:00000000-0000-0000-0000-000000000000
    options.app_key_secret = app_key.c_str();
   
    // the number of threads to use for orders and prices
    options.num_order_threads = 2;
    options.num_price_threads = 2;
    
    MyOrderBookHandler myOBHandler;
    SDKEventHandler sdkObserver;
    if (!ttsdk::Initialize(options, &sdkObserver, &myOBHandler))
    {
        std::cout << "Unable to initialize SDK!" << std::endl;
        return -1;
    }
    
    // wait for the SDK to be initialized as ready to handle price requests
    std::unique_lock<std::mutex> lock(mutex);
    if (sdkReadyCondition.wait_for(lock, std::chrono::seconds(300)) == std::cv_status::timeout)
    {
        std::cout << "Timeout waiting for SDK to initialize!" << std::endl;
        return -1;
    }

    //Find the Instrument 1
    ttsdk::Instrument::ResponseCode respCode;
    ttsdk::MarketId market = ttsdk::MarketId::CME;
    const char* product = "GE";
    ttsdk::ProductType productType = ttsdk::ProductType::Future;
    const char* alias = "GE Mar29";

    ttsdk::InstrumentPtr instrument1 = ttsdk::GetInstrument(market, product, productType, alias, respCode);

    if (!instrument1)
    {
        std::cout << "Unable to get instrument market: " << ttsdk::ToString(market) << " product : " << product << " type : " << ttsdk::ToString(productType) << " alias : " << alias << " Stopping strategy." << std::endl;
        return -1;
    }

    //Find the Instrument 2
    market = ttsdk::MarketId::CME;
    product = "GE";
    productType = ttsdk::ProductType::Future;
    alias = "GE Jun28";

    ttsdk::InstrumentPtr instrument2 = ttsdk::GetInstrument(market, product, productType, alias, respCode);

    if (!instrument2)
    {
        std::cout << "Unable to get instrument market: " << ttsdk::ToString(market) << " product : " << product << " type : " << ttsdk::ToString(productType) << " alias : " << alias << " Stopping strategy." << std::endl;
        return -1;
    }

    //The risk is reserved for the entire app, per instrument and side. 
    PositionManager posManager;
    if (!posManager.ReserveRisk(instrument1, ttsdk::RiskSide::Buy, 15.0, 5.0))
        return -1;

    if (!posManager.ReserveRisk(instrument2, ttsdk::RiskSide::Sell, 15.0, 5.0))
        return -1;

    RouteOrder myStrategy(instrument1);
    myStrategy.Start();
    std::cout << "Enter q to quit. " << std::endl;
    std::string command;
    while (std::cin >> command)
    {
        if (command == "q")
        {
            std::cout << "Quitting...\n";
            break;
        }
    }
    bool running(true);
    while (running)
    {
        running = false;
        if (myStrategy.running_)
        {
            myStrategy.Stop();
            running = true;
        }
        if (running)
        {
            std::cout << "Waiting for running strategy " << std::endl;
            std::this_thread::sleep_for(std::chrono::milliseconds(150));
        }
    }

    //Release all risk set in the app
    posManager.ReleaseAllRisk();

    ttsdk::Shutdown();
}