#!/usr/bin/env python
import zipfile
import json
import sys
import getopt
import uuid
import requests
import pds
import globals
import io
import os
from requests.exceptions import HTTPError
import xml.etree.ElementTree as ET

import urllib3

def extract_metadata_from_xml(metadata_xml_str):
    meta = {}
    root = ET.fromstring(metadata_xml_str)

    # Read off these attributes out of the root they must exist
    meta["name"] = root.attrib.get("name")
    meta["ui_rep"] = root.attrib.get("uiRep")
    meta["entry_file"] = root.attrib.get("entryFile")
    meta["algo_version"] = root.attrib.get("version")

    # Hiding from UI components
    elems_to_hide = root.find("{http://www.fixprotocol.org/FIXatdl-1-1/Core}HideFromUIComponents")
    if elems_to_hide is not None:
        hide_from = []

        for e in elems_to_hide:
            hide_from.append(e.text)

        meta["hide_from"] = hide_from

    return meta

def upload_zip_file(pkg_file, metadata_xml_str, algo_type, access_token, upload_user, onBehalfOf, instr_id=None):
    algo_res_id =  str(uuid.uuid4())
    meta_res_id = str(uuid.uuid4())

    # Upload the big zip file, it's the algo
    meta = {
        "resourceType": "customAlgoLibrary",
        "content": open(pkg_file, "rb")
    }
    pds.upload_resource("ZIP to PDS", algo_res_id, meta, access_token, onBehalfOf)


    # Upload the metadata xml file
    meta = {
        "resourceType": "text/xml",
        "content": metadata_xml_str
    }
    try:
        pds.upload_resource("metadata to PDS", meta_res_id, meta, access_token, onBehalfOf)
    except HTTPError:
        pds.delete_resource(algo_res_id, access_token)
        raise

    # Create an instrument
    requestId = "ttsdkalgo-script--{0}".format(uuid.uuid4())
    url = globals.get_pds_url() + "api/1/algoinstruments" + "?request_id=" + requestId + "&onbehalfOf=" + onBehalfOf
    headers = {
        "Authorization": "Bearer " + access_token,
        "content-type": "application/json"
    }
    metadata = extract_metadata_from_xml(metadata_xml_str)
    metadata["algo_res_id"] = algo_res_id
    metadata["meta_res_id"] = meta_res_id
    metadata["algo_type"] = algo_type
    algoRecord = {
    "ATDLTemplate"  : { "metadata" : meta_res_id},
    "def"           : { "metadata" : algo_res_id},
    "algoVersion"   : metadata["algo_version"],
    "algoType"  : 12,
    "ownerId"   : upload_user,
    "name"      : metadata["name"],
    "isDeployed"    : "true"
    }
    instr_data = {
        "n": metadata["name"],
        "a": metadata["ui_rep"],
        "od": json.dumps(metadata),
        "uui": str(upload_user)
    }

    if instr_id is not None:
        instr_data["i"] = instr_id
    instr_data["m"] = 1005
    algoRecord["algoId"] = instr_id
    instr_data["ad"] = algoRecord
    data = {"instruments": [instr_data]}
    data_str = json.dumps(data)
    print(data_str)

    print("Creating algo instrument...")
    try:
        resp = requests.post(url, headers=headers, data=data_str)
        print("")
        print(resp.content)
        print("")
        resp.raise_for_status()

        if "hide_from" in metadata:
            print("NOTE: Per metadata, algo will be hidden from: " + str(metadata["hide_from"]))
            print("")

    except HTTPError:
        pds.delete_resource(meta_res_id, access_token)
        raise

    c = json.loads(resp.content)
    i = c["instruments"][0]
    return i["i"], i["iv"]

def print_usage():
    print("")
    print(" _______   ______    __   __     _______  ___      _______  _______ ")
    print("|   ____| |      |  |  |_|  |   |   _   ||   |    |       ||       |")
    print("|  |      |  _    | |      |    |  |_|  ||   |    |    ___||   _   |")
    print("|  |____  | | |   | |     |     |       ||   |    |   | __ |  | |  |")
    print("| ____  | | |_|   | |     |     |       ||   |___ |   ||  ||  |_|  |")
    print(" _____| | |       | |   _  |    |   _   ||       ||   |_| ||       |")
    print("|_______| |______|  |__| |__|   |__| |__||_______||_______||_______|")
    print("       ______   _______  _______  ___      _______  __   __   ")
    print("      |      | |       ||       ||   |    |       ||  | |  |  ")
    print("      |  _    ||    ___||    _  ||   |    |   _   ||  |_|  |  ")
    print("      | | |   ||   |___ |   |_| ||   |    |  | |  ||       |  ")
    print("      | |_|   ||    ___||    ___||   |___ |  |_|  ||_     _|  ")
    print("      |       ||   |___ |   |    |       ||       |  |   |    ")
    print("      |______| |_______||___|    |_______||_______|  |___|    ")
    print("")
    print('Usage:')
    print('deploy.py -f <algo.zip> -e environment -s appkey:secret')

def main(argv):
    env = ""
    appkey = ""
    file_name = ""
    algo_type = 12 # 12 = SDK Algo
    instr_id = None

    try:
        opts, args = getopt.getopt(argv, "hf:s:e:", ["file=", "appkeysecret=", "env="])
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_usage()
            sys.exit()
        elif opt in ("-f", "--file"):
            file_name = arg
        elif opt in ("-s", "--appkeysecret"):
            appkey = arg
        elif opt in ("-e", "--env"):
            env = arg

    if file_name.strip() == "":
        print_usage()
        sys.exit(2)

    globals.environment(env)
    theRequestId = "ttsdkalgo-deploy--{0}".format(uuid.uuid4())

    print("Obtaining access token")
    token = globals.get_access_token(appkey, theRequestId)
    upload_user = globals.get_user_id(token, theRequestId)
    print(str(upload_user) + ", " + token)

    print("Loading metadata data")
    f = open(file_name)
    metadata_xml_str = f.read()

    print("Creating temporary zip file to upload")
    zipFilename = file_name.replace('.xml', '.zip')
    zipFile = zipfile.ZipFile(zipFilename, 'w', zipfile.ZIP_DEFLATED, False)
    zipFile.writestr('metadata.xml', metadata_xml_str)
    zipFile.close()

    # Check if this algo is already there
    metadata = extract_metadata_from_xml(metadata_xml_str)
    algos = pds.get_algo_list(token, theRequestId)
    for algo_instr in algos["instruments"]:
        if algo_instr["n"] == metadata["name"]:
            instr_id = algo_instr["i"]
            print("Found instrument {0} with same name ({1}). Updating.".format(
                instr_id, algo_instr["n"]))
            break

    print("Uploading algo resources")
    i, iv = upload_zip_file(zipFilename, metadata_xml_str, algo_type, token, upload_user, theRequestId, instr_id)
    os.remove(zipFilename)

    print("")
    print("Successfully deployed sdk algo: InstrumentId: {0}, ver: {1}".format(i, iv))
    print("")

if __name__ == "__main__":
    main(sys.argv[1:])
